# vim:set et sts=4 sw=4:
# -*- coding: utf-8 -*-
#
# ibus - The Input Bus
#
# Copyright (c) 2015 Peng Huang <shawn.p.huang@gmail.com>
# Copyright (c) 2015-2023 Takao Fujiwara <takao.fujiwara1@gmail.com>
# Copyright (c) 2013-2015 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

# This file is ported from
# gnome-control-center/panels/region/cc-input-chooser.c

from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import IBus

import functools
import gettext
import i18n
import locale

from icon import load_icon
from i18n import _, N_

ROW_TRAVEL_DIRECTION_NONE,      \
ROW_TRAVEL_DIRECTION_FORWARD,   \
ROW_TRAVEL_DIRECTION_BACKWARD = list(range(3))

class EngineDialog(Gtk.Dialog):
    __gtype_name__ = 'EngineDialog'
    __initial_languages = [ IBus.get_language_name('en_US'),
                            IBus.get_language_name('en_GB'),
                            IBus.get_language_name('de_DE'),
                            IBus.get_language_name('fr_FR'),
                            IBus.get_language_name('es_ES'),
                            IBus.get_language_name('zh_CN'),
                            IBus.get_language_name('ja_JP'),
                            IBus.get_language_name('ru_RU'),
                            IBus.get_language_name('ar_EG') ]


    def __init__(self, transient_for = None):
        super(EngineDialog, self).__init__(
                title = _("Select an input method"),
                transient_for = transient_for,
                resizable = True)
        buttons = (_("_Cancel"), Gtk.ResponseType.CANCEL,
                   _("_Add"), Gtk.ResponseType.APPLY)
        self.add_buttons(*buttons)
        self.set_response_sensitive(Gtk.ResponseType.APPLY, False)

        self.__engines_for_lang = {}
        self.__untrans_for_lang = {}
        self.__langs = {}

        self.__scrolled = Gtk.ScrolledWindow(
                hscrollbar_policy = Gtk.PolicyType.NEVER,
                vscrollbar_policy = Gtk.PolicyType.NEVER,
                shadow_type = Gtk.ShadowType.IN,
                margin_start = 6,
                margin_end = 6,
                margin_top = 6,
                margin_bottom = 6)
        self.vbox.add(self.__scrolled)
        viewport = Gtk.Viewport()
        self.__scrolled.add(viewport)
        self.__list = Gtk.ListBox(vexpand = True,
                                  halign = Gtk.Align.FILL,
                                  valign = Gtk.Align.FILL)
        viewport.add(self.__list)

        self.__adjustment = self.__scrolled.get_vadjustment()
        self.__list.set_adjustment(self.__adjustment)
        self.__list.set_filter_func(self.__list_filter, None)
        self.__list.connect('row-activated', self.__row_activated)
        self.__list.connect('row-selected', self.__row_selected)

        self.__showing_extra = False
        self.__more_row = self.__more_row_new()

        self.__filter_timeout_id = 0
        self.__filter_word = None
        self.__filter_entry = Gtk.SearchEntry(hexpand = True,
                                              margin_start = 6,
                                              margin_end = 6,
                                              margin_top = 6,
                                              margin_bottom = 6)
        self.__filter_entry.set_no_show_all(True)
        self.__filter_entry.connect('search-changed', self.__filter_changed)
        self.vbox.add(self.__filter_entry)

        self.show_all()


    def __list_filter(self, row, data):
        if row == self.__more_row:
            return not self.__showing_extra
        if not self.__showing_extra and row.is_extra:
            return False
        if self.__filter_word == None:
            return True
        if row.back:
            return True

        words = self.__filter_word.lower().strip().split()
        if all(word in row.name.lower() for word in words):
            return True
        if all(word in row.untrans.lower() for word in words):
            return True
        # Search engine name in language list
        if row.lang_info:
            if row.name in self.__engines_for_lang.keys():
                for row_l in self.__engines_for_lang[row.name]:
                    if all(word in row_l.name.lower() for word in words):
                        return True
                    if all(word in row_l.untrans.lower() for word in words):
                        return True
        # Search language name in engine list
        if not row.lang_info:
            for l in self.__engines_for_lang.keys():
                if all(word in l.lower() for word in words):
                    for row_l in self.__engines_for_lang[l]:
                        if row.name == row_l.name:
                            return True
            for (trans, untrans) in self.__untrans_for_lang.items():
                if all(word in untrans.lower() for word in words):
                    for row_l in self.__engines_for_lang[trans]:
                        if row.name == row_l.name:
                            return True
        return False


    def __row_activated(self, box, row):
        if row == self.__more_row:
            # Undo sensitive Gtk.ResponseType.APPLY button
            self.__list.unselect_row(row)
            self.__show_more()
            return
        if row.back:
            self.__show_lang_rows()
            return
        if row.lang_info:
            self.__show_engines_for_lang(row)
            return


    def __row_selected(self, box, row):
        self.set_response_sensitive(Gtk.ResponseType.APPLY, row != None)


    def __padded_label_new(self, text, icon, alignment, direction):
        hbox = Gtk.Box(orientation = Gtk.Orientation.HORIZONTAL)

        if direction == ROW_TRAVEL_DIRECTION_BACKWARD:
            rtl = (Gtk.Widget.get_default_direction() == \
                   Gtk.TextDirection.RTL)
            if rtl:
                arrow = Gtk.Image.new_from_icon_name(
                    'go-previous-rtl-symbolic', Gtk.IconSize.MENU)
            else:
                arrow = Gtk.Image.new_from_icon_name(
                    'go-previous-symbolic', Gtk.IconSize.MENU)
            hbox.pack_start(arrow, False, True, 0)

        if icon != None:
            pixbuf = load_icon(icon, Gtk.IconSize.LARGE_TOOLBAR)
            image = Gtk.Image(pixbuf = pixbuf)
            hbox.pack_start(image, False, True, 0)

        label = Gtk.Label(label = text)
        label.set_halign(alignment)
        label.set_valign(Gtk.Align.CENTER)
        label.set_margin_start(20)
        label.set_margin_end(20)
        label.set_margin_top(6)
        label.set_margin_bottom(6)
        hbox.pack_start(label, True, True, 0)
        return hbox


    def __list_box_row_new(self, text):
        row = Gtk.ListBoxRow()
        row.name = text
        row.is_extra = False
        row.lang_info = False
        row.back = False
        row.untrans = ''
        row.engine = None
        return row


    def __lang_row_new(self, text):
        row = self.__list_box_row_new(text)
        row.lang_info = True
        if len(self.__untrans_for_lang) != 0:
            row.untrans = self.__untrans_for_lang[text]
        if not self.__showing_extra and text not in self.__initial_languages:
            row.is_extra = True
        widget = self.__padded_label_new(text,
                                         None,
                                         Gtk.Align.CENTER,
                                         ROW_TRAVEL_DIRECTION_NONE)
        row.add(widget)
        return row


    def __more_row_new(self):
        row = Gtk.ListBoxRow()
        hbox = Gtk.Box(orientation = Gtk.Orientation.HORIZONTAL)
        row.add(hbox)
        row.set_tooltip_text(_("More…"))
        arrow = Gtk.Image.new_from_icon_name('view-more-symbolic',
                                             Gtk.IconSize.MENU)
        arrow.set_margin_start(20)
        arrow.set_margin_end(20)
        arrow.set_margin_top(6)
        arrow.set_margin_bottom(6)
        arrow.set_halign(Gtk.Align.CENTER)
        arrow.set_valign(Gtk.Align.CENTER)
        hbox.pack_start(arrow, True, True, 0)
        return row


    def __back_row_new(self, text):
        row = self.__list_box_row_new(text)
        row.lang_info = True
        row.back = True
        widget = self.__padded_label_new(text,
                                         None,
                                         Gtk.Align.CENTER,
                                         ROW_TRAVEL_DIRECTION_BACKWARD)
        row.add(widget)
        return row


    def __engine_row_new(self, engine):
        longname = i18n.gettext_engine_longname(engine)
        description = i18n.gettext_engine_description(engine)
        row = self.__list_box_row_new(longname)
        row.untrans = engine.get_longname()
        row.rank = engine.get_rank()
        row.set_tooltip_text(description)
        row.engine = engine
        widget = self.__padded_label_new(longname,
                                         engine.get_icon(),
                                         Gtk.Align.START,
                                         ROW_TRAVEL_DIRECTION_NONE)
        row.add(widget)
        return row


    def __set_fixed_size(self):
        if self.__scrolled.get_policy()[0] == Gtk.PolicyType.AUTOMATIC:
            return
        (width, height) = self.get_size()
        self.set_size_request(width, height)
        self.__scrolled.set_policy(Gtk.PolicyType.AUTOMATIC,
                                   Gtk.PolicyType.AUTOMATIC)


    def __remove_all_children(self):
        for l in self.__list.get_children():
            self.__list.remove(l)


    def __add_engine_rows_for_lang(self, row):
        lang = row.name

        def cmp_engine(a, b):
            if a.rank == b.rank:
                return locale.strcoll(a.name, b.name)
            return int(b.rank - a.rank)

        self.__engines_for_lang[lang].sort(
                key = functools.cmp_to_key(cmp_engine))
        for row in self.__engines_for_lang[lang]:
            self.__list.add(row)


    def __show_lang_rows(self):
        self.__remove_all_children()
        for lang in self.__langs:
            row = self.__lang_row_new(lang)
            self.__list.add(row)
        self.__list.add(self.__more_row)
        self.__list.show_all()
        self.__adjustment.set_value(self.__adjustment.get_lower())
        self.__list.invalidate_filter()
        self.__list.set_selection_mode(Gtk.SelectionMode.SINGLE)


    def __show_more(self):
        self.__set_fixed_size()
        self.__filter_entry.show()
        # Get focus after clicking on the view-more row:
        self.__filter_entry.grab_focus_without_selecting()
        self.__showing_extra = True
        self.__list.invalidate_filter()


    def __show_engines_for_lang(self, row):
        text = row.name
        self.__set_fixed_size()
        self.__remove_all_children()
        row = self.__back_row_new(text)
        self.__list.add(row)
        self.__add_engine_rows_for_lang(row)
        self.__list.show_all()
        self.__adjustment.set_value(self.__adjustment.get_lower())


    def __do_filter(self):
        text = self.__filter_entry.get_text()
        if text == '':
            self.__filter_word = None
        else:
            self.__filter_word = text
        self.__list.invalidate_filter()
        self.__filter_timeout_id = 0
        return False


    def __filter_changed(self, entry):
        if self.__filter_timeout_id == 0:
            self.__filter_timeout_id = GLib.timeout_add(150, self.__do_filter)


    def set_engines(self, engines):
        self.__engines_for_lang = {}
        self.__untrans_for_lang = {}
        for e in engines:
            l = IBus.get_language_name(e.get_language())
            if l == None:
                l = ''
            if l not in self.__engines_for_lang:
                self.__engines_for_lang[l] = []
            i18n.init_textdomain(e.get_textdomain())
            row = self.__engine_row_new(e)
            self.__engines_for_lang[l].append(row)

            # Retrieve Untranslated language names.
            untrans = IBus.get_untranslated_language_name(e.get_language())
            if untrans == None:
                untrans = ''
            self.__untrans_for_lang[l] = untrans

        keys = list(self.__engines_for_lang.keys())
        keys.sort(key=functools.cmp_to_key(locale.strcoll))
        loc = locale.getlocale()[0]
        # None on C locale
        if loc == None or loc == 'C':
            loc = 'en_US'
        current_lang = IBus.get_language_name(loc)
        # move current language to the first place
        if current_lang in keys:
            keys.remove(current_lang)
            keys.insert(0, current_lang)

        # move English to the second place
        en_lang = IBus.get_language_name('en_US')
        if en_lang != current_lang and en_lang in keys:
            keys.remove(en_lang)
            keys.insert(1, en_lang)

        #add 'Others' to the end of the combo box
        if IBus.get_language_name('Other') in keys:
            keys.remove(IBus.get_language_name('Other'))
            keys += [IBus.get_language_name('Other')]

        self.__langs = keys
        self.__show_lang_rows()


    def get_selected_engine(self):
        row = self.__list.get_selected_row()
        if row == None:
            return None
        return row.engine



# vim:set et sts=4 sw=4:
#
# ibus - The Input Bus
#
# Copyright(c) 2007-2015 Peng Huang <shawn.p.huang@gmail.com>
# Copyright(c) 2012-2021 Takao Fujiwara <takao.fujiwara1@gmail.com>
# Copyright(c) 2007-2015 Google, Inc.
# Copyright(c) 2012-2021 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

import locale
import gettext
import os
import sys

DOMAINNAME = "ibus10"

_ = lambda a: gettext.dgettext(DOMAINNAME, a)
N_ = lambda a: a

PY3K = sys.version_info >= (3, 0)
LOCALEDIR = os.getenv("IBUS_LOCALEDIR")

def init_textdomain(domainname):
    if domainname == '':
        return
    # Python's locale module doesn't provide all methods on some
    # operating systems like FreeBSD
    try:
        locale.bindtextdomain(domainname, LOCALEDIR)
        locale.bind_textdomain_codeset(domainname, 'UTF-8')
    except AttributeError:
        pass
    gettext.bindtextdomain(domainname, LOCALEDIR)
    # https://docs.python.org/3/library/gettext.html#gettext.lgettext
    if not PY3K:
        gettext.bind_textdomain_codeset(domainname, 'UTF-8')

def gettext_engine_longname(engine):
    name = engine.get_name()
    if (name.startswith('xkb:')):
        return gettext.dgettext('xkeyboard-config', engine.get_longname())
    textdomain = engine.get_textdomain()
    if textdomain == '':
        return engine.get_longname()
    return gettext.dgettext(textdomain, engine.get_longname())

def gettext_engine_description(engine):
    name = engine.get_name()
    if (name.startswith('xkb:')):
        return gettext.dgettext('xkeyboard-config', engine.get_description())
    textdomain = engine.get_textdomain()
    if textdomain == '':
        return engine.get_description()
    return gettext.dgettext(textdomain, engine.get_description())



# vim:set et sts=4 sw=4:
#
# ibus - The Input Bus
#
# Copyright (c) 2007-2015 Peng Huang <shawn.p.huang@gmail.com>
# Copyright (c) 2014-2020 Takao Fujiwara <takao.fujiwara1@gmail.com>
# Copyright (c) 2007-2020 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import IBus
from gi.repository import Pango

import i18n

from icon import load_icon
from i18n import _, N_

class EngineTreeView(Gtk.TreeView):
    __gtype_name__ = 'EngineTreeView'
    __gproperties__ = {
        'active-engine' : (
            object,
            'selected engine',
            'selected engine',
            GObject.ParamFlags.READABLE),
        'engines' : (
            object,
            'engines',
            'engines',
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE)
    }

    def __init__(self):
        super(EngineTreeView, self).__init__()

        self.__engines = []
        self.__changed = False

        # self.set_headers_visible(True)
        self.set_reorderable(True)

        self.__model = Gtk.ListStore(GObject.TYPE_PYOBJECT, GObject.TYPE_STRING)
        self.set_model(self.__model)
        self.__model.connect("row-changed", self.__emit_changed_delay_cb, "row-changed")
        self.__model.connect("row-deleted", self.__emit_changed_delay_cb, "row-deleted")
        self.__model.connect("row-inserted", self.__emit_changed_delay_cb, "row-inserted")
        self.__model.connect("rows-reordered", self.__emit_changed_delay_cb, "rows-reordered")
        self.__model.set_default_sort_func(self.__sort_engines, None)
        self.__model.set_sort_column_id(-1, Gtk.SortType.ASCENDING)

        # create im name & icon column
        column = Gtk.TreeViewColumn(_("Input Method"))
        column.set_min_width(220)

        renderer = Gtk.CellRendererPixbuf()
        renderer.set_property("xalign", 0)
        column.pack_start(renderer, False)
        column.set_cell_data_func(renderer, self.__icon_cell_data_cb)

        renderer = Gtk.CellRendererText()
        renderer.set_property("xalign", 0)
        renderer.set_property("ellipsize", Pango.EllipsizeMode.END)
        column.pack_start(renderer, True)
        column.set_cell_data_func(renderer, self.__name_cell_data_cb)
        self.append_column(column)

        # create im keyboard layout column
        renderer = Gtk.CellRendererCombo()
        model = Gtk.ListStore(GObject.TYPE_STRING)
        model.append(("us",))
        model.append(("jp",))
        model.append(("xkb",))
        renderer.set_property("xalign", 0)
        renderer.set_property("model", model)
        renderer.set_property("text-column", 0)
        renderer.set_property("has-entry", False)
        renderer.set_property("editable", True)

        column = Gtk.TreeViewColumn("Kbd")
        column.set_expand(False)
        column.set_fixed_width(32)
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        column.pack_start(renderer, False)
        column.set_cell_data_func(renderer, self.__layout_cell_data_cb)
        # self.append_column(column)

        self.get_selection().connect("changed", self.__selection_changed_cb)

    def __sort_engines(self, model, a, b, data):
        engine_a = model[a][0]
        engine_b = model[b][0]
        language_a = IBus.get_language_name(engine_a.get_language())
        language_b = IBus.get_language_name(engine_b.get_language())
        longname_a = i18n.gettext_engine_longname(engine_a)
        longname_b = i18n.gettext_engine_longname(engine_b)
        label_a = "%s - %s" % (language_a, longname_a)
        label_b = "%s - %s" % (language_b, longname_b)
        # http://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
        return (label_a > label_b) - (label_a < label_b)

    def __selection_changed_cb(self, *args):
        self.notify("active-engine");

    def __emit_changed(self, *args):
        if self.__changed:
            self.__changed = False
            self.notify("engines")

    def __emit_changed_delay_cb(self, *args):
        if not self.__changed:
            self.__changed = True
            GLib.idle_add(self.__emit_changed)


    def __icon_cell_data_cb(self, celllayout, renderer, model, it, data):
        engine = self.__model.get_value(it, 0)

        # When append_engine() is called, self.__model.append(None)
        # is called internally and engine == None could happen in
        # a slow system.
        if engine == None:
            return

        icon_size = Gtk.icon_size_lookup(Gtk.IconSize.LARGE_TOOLBAR)[0]
        pixbuf = load_icon(engine.get_icon(), Gtk.IconSize.LARGE_TOOLBAR)
        renderer.set_property("pixbuf", pixbuf)

    def __name_cell_data_cb(self, celllayout, renderer, model, it, data):
        engine = self.__model.get_value(it, 0)

        # When append_engine() is called, self.__model.append(None)
        # is called internally and engine == None could happen in
        # a slow system.
        if engine == None:
            return

        renderer.set_property("sensitive", True)
        language = IBus.get_language_name(engine.get_language())
        longname = i18n.gettext_engine_longname(engine)
        renderer.set_property("text",
                "%s - %s" % (language, longname))
        renderer.set_property("weight", Pango.Weight.NORMAL)

    def __layout_cell_data_cb(self, celllayout, renderer, model, it, data):
        engine = self.__model.get_value(it, 0)
        layout = self.__model.get_value(it, 1)
        renderer.set_property("sensitive", True)
        if not layout:
            layout = engine.layout
        renderer.set_property("text", layout)
        renderer.set_property("weight", Pango.Weight.NORMAL)

    def do_get_property(self, prop):
        if prop.name == "active-engine":
            it = self.get_selected_iter()
            if it == None:
                return None
            row = self.__model.get(it, 0)
            return row[0]
        elif prop.name == "engines":
            engines = [ r[0] for r in self.__model if r[0] != None]
            return engines
        else:
            raise AttributeError('unknown property %s' % prop.name)

    def do_set_property(self, prop, value):
        if prop.name == "active-engine":
            raise AttributeError("active-engine is readonly")
        elif prop.name == "engines":
            set_engines(value)
        else:
            raise AttributeError('unknown property %s' % prop.name)

    def set_engines(self, engines):
        self.__model.clear()
        self.__engines = []
        for e in engines:
            if e in self.__engines:
                continue
            it = self.__model.append(None)
            i18n.init_textdomain(e.get_textdomain())
            self.__model.set(it, 0, e)
            self.__engines.append(e)
        self.__emit_changed()

    def get_selected_iter(self):
        selection = self.get_selection()
        if selection:
            return selection.get_selected()[1]

    def get_engines(self):
        return self.__engines

    def get_sorted_engines(self):
        return self.get_property("engines")

    def get_active_engine(self):
        return self.get_property("active-engine")

    def prepend_engine(self, engine):
        if engine == None or engine in self.__engines:
            return
        it = self.__model.prepend(None)
        self.__model.set(it, 0, engine)
        self.__engines = [engine] + self.__engines

    def append_engine(self, engine):
        if engine == None or engine in self.__engines:
            return
        it = self.__model.append(None)
        self.__model.set(it, 0, engine)
        self.__engines.append(engine)

    def remove_engine(self):
        it = self.get_selected_iter()
        if it == None:
            return
        row = self.__model[it]
        engine = row[0]
        self.__engines.remove(engine)
        index = row.path.get_indices()[0]
        self.__model.remove(it)
        try:
            row = self.__model[index]
            selection = self.get_selection()
            selection.select_path(row.path)
        except:
            pass

    def move_up_engine(self):
        it = self.get_selected_iter()
        if it == None:
            return
        row = self.__model[it]
        index = row.path.get_indices()[0]
        if index == 0:
            return
        self.__model.swap(it, self.__model[index - 1].iter)
        self.scroll_to_cell(row.path, None)

    def move_down_engine(self):
        it = self.get_selected_iter()
        if it == None:
            return
        row = self.__model[it]
        index = row.path.get_indices()[0]
        last_row = self.__model[-1]
        last_index = last_row.path.get_indices()[0]
        if index == last_index:
            return
        self.__model.swap(it, self.__model[index + 1].iter)
        self.scroll_to_cell(row.path, None)

GObject.type_register(EngineTreeView)

if __name__ == "__main__":
    tree = EngineTreeView()
    tree.set_engines([IBus.EngineDesc(language="zh")])
    w = Gtk.Window()
    w.add(tree)
    w.show_all()
    Gtk.main()



# vim:set et sts=4 sw=4:
# -*- coding: utf-8 -*-
#
# ibus - The Input Bus
#
# Copyright (c) 2007-2016 Peng Huang <shawn.p.huang@gmail.com>
# Copyright (c) 2010-2020 Takao Fujiwara <takao.fujiwara1@gmail.com>
# Copyright (c) 2007-2020 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA

# for python2
from __future__ import print_function

import os
import signal
import sys
import time
import glob

from gi import require_version as gi_require_version
gi_require_version('GLib', '2.0')
gi_require_version('GdkX11', '3.0')
gi_require_version('Gio', '2.0')
gi_require_version('Gtk', '3.0')
gi_require_version('IBus', '1.0')

from gi.repository import GLib
# set_prgname before importing other modules to show the name in warning
# messages when import modules are failed.
GLib.set_prgname('ibus-setup')

from gi.repository import GdkX11
from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import IBus
from os import path

import i18n
import keyboardshortcut
import locale
from emojilang import EmojiLangButton
from enginecombobox import EngineComboBox
from enginedialog import EngineDialog
from enginetreeview import EngineTreeView
from engineabout import EngineAbout
from i18n import DOMAINNAME, _, N_

(
    COLUMN_NAME,
    COLUMN_ENABLE,
    COLUMN_PRELOAD,
    COLUMN_VISIBLE,
    COLUMN_ICON,
    COLUMN_DATA,
) = list(range(6))

(
    DATA_NAME,
    DATA_LOCAL_NAME,
    DATA_LANG,
    DATA_ICON,
    DATA_AUTHOR,
    DATA_CREDITS,
    DATA_EXEC,
    DATA_STARTED,
    DATA_PRELOAD
) = list(range(9))

class Setup(object):
    def __flush_gtk_events(self):
        while Gtk.events_pending():
            Gtk.main_iteration()

    def __init__(self):
        super(Setup, self).__init__()

        self.__settings_general = Gio.Settings(
                schema = "org.freedesktop.ibus.general");
        self.__settings_hotkey = Gio.Settings(
                schema = "org.freedesktop.ibus.general.hotkey");
        self.__settings_panel = Gio.Settings(
                schema = "org.freedesktop.ibus.panel");
        self.__settings_emoji = Gio.Settings(
                schema = "org.freedesktop.ibus.panel.emoji");

        # IBus.Bus() calls ibus_bus_new().
        # Gtk.Builder().add_from_file() also calls ibus_bus_new_async()
        # via ibus_im_context_new().
        # Then if IBus.Bus() is called after Gtk.Builder().add_from_file(),
        # the connection delay would be happened without an async
        # finish function.
        self.__bus = None
        self.__init_bus()

        # Gtk.ListBox has been available since gtk 3.10
        self.__has_list_box = hasattr(Gtk, 'ListBox')

        gtk_builder_file = path.join(path.dirname(__file__), "./setup.ui")
        self.__builder = Gtk.Builder()
        self.__builder.set_translation_domain(DOMAINNAME)
        self.__builder.add_from_file(gtk_builder_file);
        self.__init_ui()

    def __init_hotkeys(self):
        name = 'triggers'
        label = 'switch_engine'
        comment = \
            _("Use shortcut with shift to switch to the previous input method") 
        self.__init_hotkey(name, label, comment)
        name = 'emoji'
        label = 'emoji_dialog'
        self.__init_hotkey(name, label)
        name = 'unicode'
        label = 'unicode_dialog'
        self.__init_hotkey(name, label)

    def __init_hotkey(self, name, label, comment=None):
        if name == 'emoji':
            shortcuts = self.__settings_emoji.get_strv('hotkey')
        elif name == 'unicode':
            shortcuts = self.__settings_emoji.get_strv('unicode-hotkey')
        else:
            shortcuts = self.__settings_hotkey.get_strv(name)
        button = self.__builder.get_object("button_%s" % label)
        entry = self.__builder.get_object("entry_%s" % label)
        entry.set_text("; ".join(shortcuts))
        tooltip = "\n".join(shortcuts)
        if comment != None:
            tooltip += "\n" + comment
        entry.set_tooltip_text(tooltip)
        if name == 'emoji':
            button.connect("clicked", self.__shortcut_button_clicked_cb,
                    'hotkey', 'panel/' + name, label, entry)
        elif name == 'unicode':
            button.connect("clicked", self.__shortcut_button_clicked_cb,
                    'unicode-hotkey', 'panel/emoji', label, entry)
        else:
            button.connect("clicked", self.__shortcut_button_clicked_cb,
                    name, "general/hotkey", label, entry)

    def __init_panel(self):
        # lookup table orientation
        self.__combobox_lookup_table_orientation = self.__builder.get_object(
                "combobox_lookup_table_orientation")
        self.__settings_panel.bind('lookup-table-orientation',
                                   self.__combobox_lookup_table_orientation,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)

        # auto hide
        self.__combobox_panel_show = self.__builder.get_object(
                "combobox_panel_show")
        self.__settings_panel.bind('show',
                                   self.__combobox_panel_show,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)

        # panel position
        self.__combobox_panel_position = self.__builder.get_object(
                "combobox_panel_position")
        self.__combobox_panel_position.set_active(3)
        #self.__settings_panel.bind('position',
        #                           self.__combobox_panel_position,
        #                           'active',
        #                           Gio.SettingsBindFlags.DEFAULT)

        # custom font
        self.__checkbutton_custom_font = self.__builder.get_object(
                "checkbutton_custom_font")
        self.__settings_panel.bind('use-custom-font',
                                   self.__checkbutton_custom_font,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)

        self.__fontbutton_custom_font = self.__builder.get_object(
                "fontbutton_custom_font")
        self.__settings_panel.bind('custom-font',
                                    self.__fontbutton_custom_font,
                                   'font-name',
                                   Gio.SettingsBindFlags.DEFAULT)
        self.__settings_panel.bind('use-custom-font',
                                    self.__fontbutton_custom_font,
                                   'sensitive',
                                   Gio.SettingsBindFlags.GET)

        # custom theme
        self.__model_custom_theme = self.__builder.get_object(
                "model_custom_theme")
        self.__combobox_custom_theme = self.__builder.get_object(
                "combobox_custom_theme")
        self.__checkbutton_custom_theme = self.__builder.get_object(
                "checkbutton_custom_theme")

        def update_combobox_custom_theme(settings, key):
            theme_name_list = self.__init_available_gtk_themes()
            self.__model_custom_theme.clear()
            for name in theme_name_list:
                self.__model_custom_theme.append([name])
            current_theme = self.__settings_panel.get_string(key)
            try:
                current_theme_number = theme_name_list.index(current_theme)
            except ValueError:
                self.__settings_panel.reset(key)
                current_theme = self.__settings_panel.get_string(key)
                current_theme_number = theme_name_list.index(current_theme)
            self.__combobox_custom_theme.set_active(current_theme_number)

        update_combobox_custom_theme(None, 'custom-theme')
        self.__settings_panel.bind('use-custom-theme',
                                   self.__checkbutton_custom_theme,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)
        self.__settings_panel.connect('changed::custom-theme',
                                   update_combobox_custom_theme)
        self.__settings_panel.bind('use-custom-theme',
                                   self.__combobox_custom_theme,
                                   'sensitive',
                                   Gio.SettingsBindFlags.DEFAULT)
        self.__combobox_custom_theme.connect("changed",
                                   self.__on_combobox_custom_theme_changed)


        # custom icon
        self.__model_custom_icon = self.__builder.get_object(
                "model_custom_icon")
        self.__combobox_custom_icon = self.__builder.get_object(
                "combobox_custom_icon")
        self.__checkbutton_custom_icon = self.__builder.get_object(
                "checkbutton_custom_icon")

        def update_combobox_custom_icon(settings, key):
            icon_name_list = self.__init_available_gtk_icons()
            self.__model_custom_icon.clear()
            for name in icon_name_list:
                self.__model_custom_icon.append([name])
            current_icon = self.__settings_panel.get_string(key)
            try:
                current_icon_number = icon_name_list.index(current_icon)
            except ValueError:
                self.__settings_panel.reset(key)
                current_icon = self.__settings_panel.get_string(key)
                current_icon_number = icon_name_list.index(current_icon)
            self.__combobox_custom_icon.set_active(current_icon_number)

        update_combobox_custom_icon(None, 'custom-icon')
        self.__settings_panel.bind('use-custom-icon',
                                   self.__checkbutton_custom_icon,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)
        self.__settings_panel.connect('changed::custom-icon',
                                   update_combobox_custom_icon)
        self.__settings_panel.bind('use-custom-icon',
                                   self.__combobox_custom_icon,
                                   'sensitive',
                                   Gio.SettingsBindFlags.DEFAULT)
        self.__combobox_custom_icon.connect("changed",
                                   self.__on_combobox_custom_icon_changed)

        # show icon on system tray
        self.__checkbutton_show_icon_on_systray = self.__builder.get_object(
                "checkbutton_show_icon_on_systray")
        self.__settings_panel.bind('show-icon-on-systray',
                                   self.__checkbutton_show_icon_on_systray,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)

        # show ime name
        self.__checkbutton_show_im_name = self.__builder.get_object(
                "checkbutton_show_im_name")
        self.__settings_panel.bind('show-im-name',
                                   self.__checkbutton_show_im_name,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)

        self.__checkbutton_glyph_from_engine_lang = self.__builder.get_object(
                "checkbutton_use_glyph_from_engine_lang")
        self.__settings_panel.bind('use-glyph-from-engine-lang',
                                   self.__checkbutton_glyph_from_engine_lang,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)

    def __init_general(self):
        # embed preedit text
        self.__checkbutton_embed_preedit_text = self.__builder.get_object(
                "checkbutton_embed_preedit_text")
        self.__settings_general.bind('embed-preedit-text',
                                    self.__checkbutton_embed_preedit_text,
                                    'active',
                                    Gio.SettingsBindFlags.DEFAULT)

        # use system keyboard layout setting
        self.__checkbutton_use_sys_layout = self.__builder.get_object(
                "checkbutton_use_sys_layout")
        self.__settings_general.bind('use-system-keyboard-layout',
                                    self.__checkbutton_use_sys_layout,
                                    'active',
                                    Gio.SettingsBindFlags.DEFAULT)

        # use global ime setting
        self.__checkbutton_use_global_engine = self.__builder.get_object(
                "checkbutton_use_global_engine")
        self.__settings_general.bind('use-global-engine',
                                    self.__checkbutton_use_global_engine,
                                    'active',
                                    Gio.SettingsBindFlags.DEFAULT)

        # init engine page
        self.__engines = self.__bus.list_engines()
        self.__combobox = self.__builder.get_object("combobox_engines")
        if self.__has_list_box:
            self.__combobox.set_no_show_all(True)
            self.__combobox.hide()
        else:
            self.__combobox.set_engines(self.__engines)

        tmp_dict = {}
        for e in self.__engines:
            tmp_dict[e.get_name()] = e
        engine_names = self.__settings_general.get_strv('preload-engines')
        engines = [tmp_dict[name] for name in engine_names if name in tmp_dict]

        self.__treeview = self.__builder.get_object("treeview_engines")
        self.__treeview.set_engines(engines)

        button = self.__builder.get_object("button_engine_add")
        if self.__has_list_box:
            button.set_sensitive(True)
            button.connect("clicked", self.__button_engine_add_cb)
        else:
            button.connect("clicked", self.__button_engine_add_cb_deprecate)
        button = self.__builder.get_object("button_engine_remove")
        button.connect("clicked", lambda *args:self.__treeview.remove_engine())
        button = self.__builder.get_object("button_engine_up")
        button.connect("clicked", lambda *args:self.__treeview.move_up_engine())

        button = self.__builder.get_object("button_engine_down")
        button.connect("clicked",
                lambda *args:self.__treeview.move_down_engine())

        button = self.__builder.get_object("button_engine_about")
        button.connect("clicked", self.__button_engine_about_cb)

        self.__engine_setup_exec_list = {}
        button = self.__builder.get_object("button_engine_preferences")
        button.connect("clicked", self.__button_engine_preferences_cb)

        self.__combobox.connect("notify::active-engine",
                self.__combobox_notify_active_engine_cb)
        self.__treeview.connect("notify::active-engine", self.__treeview_notify_cb)
        self.__treeview.connect("notify::engines", self.__treeview_notify_cb)

    def __init_emoji(self):
        self.__fontbutton_emoji_font = self.__builder.get_object(
                'fontbutton_emoji_font')
        self.__fontbutton_emoji_font.set_preview_text('🙂🍎🚃💓📧⚽🐳');
        self.__settings_emoji.bind('font',
                                    self.__fontbutton_emoji_font,
                                   'font-name',
                                   Gio.SettingsBindFlags.DEFAULT)
        self.__button_emoji_lang = self.__builder.get_object(
                'button_emoji_lang')
        self.__settings_emoji.bind('lang',
                                    self.__button_emoji_lang,
                                   'lang',
                                   Gio.SettingsBindFlags.DEFAULT)
        self.__checkbutton_emoji_partial_match = self.__builder.get_object(
                'checkbutton_emoji_partial_match')
        checkbutton_label = self.__checkbutton_emoji_partial_match.get_child()
        if type(checkbutton_label) == Gtk.Label:
            checkbutton_label.set_property('wrap', True)
            checkbutton_label.set_property('max-width-chars', 74)
        self.__spinbutton_emoji_partial_match = self.__builder.get_object(
                'spinbutton_emoji_partial_match')
        self.__settings_emoji.bind('has-partial-match',
                                   self.__checkbutton_emoji_partial_match,
                                   'active',
                                   Gio.SettingsBindFlags.DEFAULT)
        self.__settings_emoji.bind('has-partial-match',
                                   self.__spinbutton_emoji_partial_match,
                                   'sensitive',
                                   Gio.SettingsBindFlags.GET)

        def adjustment_value_changed_cb(obj):
            key = 'partial-match-length'
            value = int(adjustment.get_value())
            if value == self.__settings_emoji.get_int(key):
                return
            self.__settings_emoji.set_int(key, value)
        def settings_emoji_partial_match_length_cb(settings, key):
            value = self.__settings_emoji.get_int(key)
            old_value = int(self.__spinbutton_emoji_partial_match.get_value())
            if value == old_value:
                return
            self.__spinbutton_emoji_partial_match.set_value(value)
        settings_emoji_partial_match_length_cb(None, 'partial-match-length')
        adjustment = self.__spinbutton_emoji_partial_match.get_adjustment()
        adjustment.connect('value-changed', adjustment_value_changed_cb)
        self.__settings_emoji.connect('changed::partial-match-length',
                                      settings_emoji_partial_match_length_cb)

        self.__hbox_emoji_partial_match = self.__builder.get_object(
                'hbox_emoji_partial_match')
        self.__settings_emoji.bind('has-partial-match',
                                   self.__hbox_emoji_partial_match,
                                   'sensitive',
                                   Gio.SettingsBindFlags.GET)
        self.__radiobutton_emoji_prefix_match = self.__builder.get_object(
                'radiobutton_emoji_prefix_match')
        self.__radiobutton_emoji_suffix_match = self.__builder.get_object(
                'radiobutton_emoji_suffix_match')
        self.__radiobutton_emoji_contain_match = self.__builder.get_object(
                'radiobutton_emoji_contain_match')

        def radiobuton_emoji_partial_match_cb(obj):
            key = 'partial-match-condition'
            condition = 0
            if not obj.get_active():
                return
            if obj == self.__radiobutton_emoji_prefix_match:
                condition = 0
            elif obj == self.__radiobutton_emoji_suffix_match:
                condition = 1
            elif obj == self.__radiobutton_emoji_contain_match:
                condition = 2
            else:
                print('Wrong emoji partial match object')
                return
            self.__settings_emoji.set_int(key, condition)
        def settings_emoji_partial_match_condition_cb(settings, key):
            value = self.__settings_emoji.get_int(key)
            obj = None
            if value == 0:
                obj = self.__radiobutton_emoji_prefix_match
            elif value == 1:
                obj = self.__radiobutton_emoji_suffix_match
            elif value == 2:
                obj = self.__radiobutton_emoji_contain_match
            else:
                print('Wrong emoji partial match condition')
                return
            if obj.get_active():
                return
            obj.set_active(True)

        settings_emoji_partial_match_condition_cb(None,
                                                  'partial-match-condition')
        self.__radiobutton_emoji_prefix_match.connect(
                'toggled',
                radiobuton_emoji_partial_match_cb)
        self.__radiobutton_emoji_suffix_match.connect(
                'toggled',
                radiobuton_emoji_partial_match_cb)
        self.__radiobutton_emoji_contain_match.connect(
                'toggled',
                radiobuton_emoji_partial_match_cb)
        self.__settings_emoji.connect('changed::partial-match-condition',
                                      settings_emoji_partial_match_condition_cb)

    def __init_ui(self):
        # add icon search path
        self.__window = self.__builder.get_object("window_preferences")
        self.__window.connect("delete-event", Gtk.main_quit)
        self.__window.connect("notify::window", self.__gdk_window_set_cb)

        self.__button_close = self.__builder.get_object("button_close")
        self.__button_close.connect("clicked", Gtk.main_quit)

        self.__init_hotkeys()
        self.__init_panel()
        self.__init_general()
        self.__init_emoji()

    def __gdk_window_set_cb(self, object, pspec):
        window = object.get_window()
        if type(window) != GdkX11.X11Window:
            return
        s = '%u' % GdkX11.X11Window.get_xid(window)
        GLib.setenv('IBUS_SETUP_XID', s, True)

    def __combobox_notify_active_engine_cb(self, combobox, property):
        engine = self.__combobox.get_active_engine()
        button = self.__builder.get_object("button_engine_add")
        button.set_sensitive(
                engine != None and engine not in self.__treeview.get_engines())

    def __get_engine_setup_exec_args(self, engine):
        args = []
        if engine == None:
           return args
        setup = str(engine.get_setup())
        if len(setup) != 0:
            args = setup.split()
            args.insert(1, path.basename(args[0]))
            return args
        name = str(engine.get_name())
        libexecdir = os.environ['IBUS_LIBEXECDIR']
        setup_path = (libexecdir + '/' + 'ibus-setup-' if libexecdir != None \
            else 'ibus-setup-') + name.split(':')[0]
        if path.exists(setup_path):
            args.append(setup_path)
            args.append(path.basename(setup_path))
        return args

    def __treeview_notify_cb(self, treeview, prop):
        if prop.name not in ("active-engine", "engines"):
            return

        engines = self.__treeview.get_sorted_engines()
        engine = self.__treeview.get_active_engine()

        self.__builder.get_object("button_engine_remove").set_sensitive(engine != None)
        self.__builder.get_object("button_engine_about").set_sensitive(engine != None)
        self.__builder.get_object("button_engine_up").set_sensitive(engine not in engines[:1])
        self.__builder.get_object("button_engine_down").set_sensitive(engine not in engines[-1:])

        obj = self.__builder.get_object("button_engine_preferences")
        if len(self.__get_engine_setup_exec_args(engine)) != 0:
            obj.set_sensitive(True)
        else:
            obj.set_sensitive(False)

        if prop.name == "engines":
            engines = self.__treeview.get_engines()
            engine_names = [e.get_name() for e in engines]
            self.__settings_general.set_strv('preload-engines', engine_names)

    def __button_engine_add_cb(self, button):
        dialog = EngineDialog(transient_for = self.__window)
        dialog.set_engines(self.__engines)
        id = dialog.run()

        if id != Gtk.ResponseType.APPLY:
            dialog.destroy()
            return

        engine = dialog.get_selected_engine()
        dialog.destroy()

        self.__treeview.append_engine(engine)

    def __button_engine_add_cb_deprecate(self, button):
        engine = self.__combobox.get_active_engine()
        self.__treeview.append_engine(engine)

    def __button_engine_about_cb(self, button):
        engine = self.__treeview.get_active_engine()
        if engine:
            about = EngineAbout(engine = engine, transient_for = self.__window)
            about.run()
            about.destroy()

    def __button_engine_preferences_cb(self, button):
        engine = self.__treeview.get_active_engine()
        args = self.__get_engine_setup_exec_args(engine)
        if len(args) == 0:
            return
        name = engine.get_name()
        if name in list(self.__engine_setup_exec_list.keys()):
            try:
                wpid, sts = os.waitpid(self.__engine_setup_exec_list[name],
                                       os.WNOHANG)
                # the setup is still running.
                if wpid == 0:
                    return
            except OSError:
                pass
            del self.__engine_setup_exec_list[name]
        self.__engine_setup_exec_list[name] = os.spawnl(os.P_NOWAIT, *args)

    def __init_bus(self):
        self.__bus = IBus.Bus()
        if self.__bus.is_connected():
            return

        message = _("The IBus daemon is not running. Do you wish to start it?")
        dlg = Gtk.MessageDialog(message_type = Gtk.MessageType.QUESTION,
                                buttons = Gtk.ButtonsType.YES_NO,
                                text = message)
        id = dlg.run()
        dlg.destroy()
        self.__flush_gtk_events()
        if id != Gtk.ResponseType.YES:
            sys.exit(0)

        main_loop = GLib.MainLoop()

        timeout = 5
        GLib.timeout_add_seconds(timeout, lambda *args: main_loop.quit())
        self.__bus.connect("connected", lambda *args: main_loop.quit())

        os.spawnlp(os.P_NOWAIT, "ibus-daemon", "ibus-daemon", "--xim", "--daemonize")

        main_loop.run()

        if self.__bus.is_connected():
            message = _("IBus has been started! "
                "If you cannot use IBus, add the following lines to your $HOME/.bashrc; then relog into your desktop.\n"
                "  export GTK_IM_MODULE=ibus\n"
                "  export XMODIFIERS=@im=ibus\n"
                "  export QT_IM_MODULE=ibus"
                )
            dlg = Gtk.MessageDialog(message_type = Gtk.MessageType.INFO,
                                    buttons = Gtk.ButtonsType.OK,
                                    text = message)
            id = dlg.run()
            dlg.destroy()
            self.__flush_gtk_events()
        else:
            # Translators: %d == 5 currently
            message = _("IBus daemon could not be started in %d seconds.")
            dlg = Gtk.MessageDialog(message_type = Gtk.MessageType.INFO,
                                    buttons = Gtk.ButtonsType.OK,
                                    text = message % timeout)
            id = dlg.run()
            dlg.destroy()
            self.__flush_gtk_events()
            sys.exit(0)

    def __shortcut_button_clicked_cb(self, button, name, section, _name, entry):
        buttons = (_("_Cancel"), Gtk.ResponseType.CANCEL,
                   _("_OK"), Gtk.ResponseType.OK)
        title1 = _("Select keyboard shortcut for %s")
        # Translators: Title of the window
        title2 = _("switching input methods")
        title = title1 % title2
        dialog = keyboardshortcut.KeyboardShortcutSelectionDialog(
                title = title, transient_for = self.__window)
        dialog.add_buttons(*buttons)
        text = entry.get_text()
        if text:
            shortcuts = text.split("; ")
        else:
            shortcuts = None
        dialog.set_shortcuts(shortcuts)
        id = dialog.run()
        shortcuts = dialog.get_shortcuts()
        dialog.destroy()
        if id != Gtk.ResponseType.OK:
            return
        if section == 'panel/emoji':
            self.__settings_emoji.set_strv(name, shortcuts)
        else:
            self.__settings_hotkey.set_strv(name, shortcuts)
        text = "; ".join(shortcuts)
        entry.set_text(text)
        tooltip = "\n".join(shortcuts)
        tooltip += "\n" + \
            _("Use shortcut with shift to switch to the previous input method") 
        entry.set_tooltip_text(tooltip)

    def __init_available_gtk_themes(self):
        path_list = []
        path_list.append(os.path.join(GLib.get_home_dir(), ".themes"))
        path_list.append(os.path.join(GLib.get_user_data_dir(), "themes"))
        path_list.extend(list(map(lambda x: os.path.join(
            x, "themes"), GLib.get_system_data_dirs())))
        theme_name_list = []
        gtk_theme_path = []
        for path in path_list:
            gtk_theme_path.extend(glob.glob(path + "/*/gtk-*/gtk.css"))
        for path in gtk_theme_path:
            theme_name_list.append(os.path.basename(
                os.path.dirname(os.path.dirname(path))))

        theme_name_list.extend([
            'Adwaita', 'HighContrast', 'HighContrastInverse'
        ])
        theme_name_list = list(set(theme_name_list))
        theme_name_list.sort()

        return theme_name_list

    def __on_combobox_custom_theme_changed(self, combobox):
        tree_iter = self.__combobox_custom_theme.get_active_iter()
        if tree_iter is not None:
            model = self.__combobox_custom_theme.get_model()
            theme_name = model[tree_iter][0]
            self.__settings_panel.set_string('custom-theme', theme_name)

    def __init_available_gtk_icons(self):
        path_list = []
        path_list.append(os.path.join(GLib.get_home_dir(), ".icons"))
        path_list.append(os.path.join(GLib.get_user_data_dir(), "icons"))
        path_list.extend(list(map(lambda x: os.path.join(
            x, "icons"), GLib.get_system_data_dirs())))
        icon_name_list = []
        gtk_icon_path = []
        for path in path_list:
            gtk_icon_path.extend(glob.glob(path + "/*/index.theme"))
        for path in gtk_icon_path:
            dir = os.path.dirname(path)
            if not os.path.exists(os.path.join(dir, "cursors")):
                icon_name_list.append(os.path.basename(dir))

        icon_name_list.extend(["Adwaita"])
        icon_name_list = list(set(icon_name_list))
        icon_name_list.sort()

        return icon_name_list

    def __on_combobox_custom_icon_changed(self, combobox):
        tree_iter = self.__combobox_custom_icon.get_active_iter()
        if tree_iter is not None:
            model = self.__combobox_custom_icon.get_model()
            icon_name = model[tree_iter][0]
            self.__settings_panel.set_string('custom-icon', icon_name)

    def __item_started_column_toggled_cb(self, cell, path_str, model):

        # get toggled iter
        iter = model.get_iter_from_string(path_str)
        data = model.get_value(iter, COLUMN_DATA)

        # do something with the value
        if data[DATA_STARTED] == False:
            try:
                self.__bus.register_start_engine(data[DATA_LANG], data[DATA_NAME])
            except Exception as e:
                dlg = Gtk.MessageDialog(message_type = Gtk.MessageType.ERROR,
                        transient_for = self.__window,
                        buttons = Gtk.ButtonsType.CLOSE,
                        text = str(e))
                dlg.run()
                dlg.destroy()
                self.__flush_gtk_events()
                return
        else:
            try:
                self.__bus.register_stop_engine(data[DATA_LANG], data[DATA_NAME])
            except Exception as e:
                dlg = Gtk.MessageDialog(message_type = Gtk.MessageType.ERROR,
                        transient_for = self.__window,
                        buttons = Gtk.ButtonsType.CLOSE,
                        text = str(e))
                dlg.run()
                dlg.destroy()
                self.__flush_gtk_events()
                return
        data[DATA_STARTED] = not data[DATA_STARTED]

        # set new value
        model.set(iter, COLUMN_ENABLE, data[DATA_STARTED])

    def __item_preload_column_toggled_cb(self, cell, path_str, model):

        # get toggled iter
        iter = model.get_iter_from_string(path_str)
        data = model.get_value(iter, COLUMN_DATA)

        data[DATA_PRELOAD] = not data[DATA_PRELOAD]
        engine = "%s:%s" % (data[DATA_LANG], data[DATA_NAME])

        if data[DATA_PRELOAD]:
            if engine not in self.__preload_engines:
                self.__preload_engines.add(engine)
                self.__settings_general.set_strv('preload-engines',
                                                 list(self.__preload_engines))
        else:
            if engine in self.__preload_engines:
                self.__preload_engines.remove(engine)
                self.__settings_general.set_strv('preload-engines',
                                                 list(self.__preload_engines))

        # set new value
        model.set(iter, COLUMN_PRELOAD, data[DATA_PRELOAD])

    def __sigusr1_cb(self, *args):
        self.__window.present()

    def run(self):
        self.__window.show_all()
        signal.signal(signal.SIGUSR1, self.__sigusr1_cb)
        Gtk.main()

if __name__ == "__main__":
    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error:
        print("Using the fallback 'C' locale", file=sys.stderr)
        locale.setlocale(locale.LC_ALL, 'C')

    i18n.init_textdomain(DOMAINNAME)
    i18n.init_textdomain('xkeyboard-config')
    setup = Setup()
    setup.run()
